/*
	gen_ellipse [n] [w] [h] [seed]
	Generate a random convex [n]-gon from a 2[w] x 2[h] ellipse.
*/

/*
	gen_few [n] [w] [h] [seed]
	Generate [n] points, then find their convex hull.
*/

/*
	gen_minkowski [n] [w] [h] [seed]
	Generate a random convex polygon with up to [n] vertices in a 2[w] x 2[h] box.
	Uses an algorithm similar to Minkowski sum.
*/

/*
	gen_polygon [n] [r] [angle] [offset] [seed]
	Generate a regular [n]-gon, inscribed in a circle of radius [r], starting from [angle] degrees.
	Then, offset each vertex by up to [offset] in each coordinate.
*/

/*
	gen_max [wh] [seed]
	Generate as much vertices as possible in a 2[wh] x 2[wh] box. Or, at least, as much as I can.
*/

#include "jngen.h"
#include <vector>
#include <string>
#include <algorithm>
using namespace std;

namespace gen_ellipse
{
	vector<pair<int, int>> half_main(int argc, char *argv[]) {
		if (argc != 5)
		{
			printf("Usage: gen_ellipse [n] [w] [h] [seed]\n");
			return {};
		}
		parseArgs(argc, argv);

		int n = getOpt(0);
		int w = getOpt(1);
		int h = getOpt(2);

		auto poly = jngen::GeometryRandom::convexPolygon(n, -w, -h, w, h);
		vector<pair<int, int>> res;
		for (auto e : poly)
			res.emplace_back(e.x, e.y);
		return res;
	}
}

namespace gen_few
{
	vector<pair<int, int>> half_main(int argc, char *argv[]) {
		if (argc != 5)
		{
			printf("Usage: gen_few [n] [w] [h] [seed]\n");
			return {};
		}
		parseArgs(argc, argv);

		int n = getOpt(0);
		int w = getOpt(1);
		int h = getOpt(2);

		auto points = jngen::TArray<jngen::Point>(n);
		for (int i = 0; i < n; ++i)
			points[i] = jngen::GeometryRandom::point(-w, -h, w, h);
		auto poly = jngen::detail::convexHull(points);
		vector<pair<int, int>> res;
		for (auto e : poly)
			res.emplace_back(e.x, e.y);
		return res;
	}
}

namespace gen_minkowski
{
	struct Point
	{
		long long x, y;
		Point() {}
		Point(long long x, long long y) : x(x), y(y) {}
		Point operator - (const Point& p) const
		{
			return Point(x - p.x, y - p.y);
		}
		Point operator + (const Point& p) const
		{
			return Point(x + p.x, y + p.y);
		}
		Point operator * (long long k) const
		{
			return Point(x * k, y * k);
		}
		long long scl(const Point& p) const
		{
			return x * p.x + y * p.y;
		}
		long long vct(const Point& p) const
		{
			return x * p.y - y * p.x;
		}
		bool operator < (const Point& p) const
		{
			return atan2(y, x) < atan2(p.y, p.x);
		}
	};

	vector<pair<int, int>> half_main(int argc, char *argv[]) {
		if (argc != 5)
		{
			printf("Usage: gen_minkowski [n] [w] [h] [seed]\n");
			return {};
		}

		int n = atoi(argv[1]);
		int w = atoi(argv[2]);
		int h = atoi(argv[3]);
		vector<Point> best_ar;
		vector<Point> ar(n);
		long long best_val = -1;
		long long pos_x, neg_x, pos_y, neg_y;
		for (int _ = 0; _ < 100; ++_)
		{
			pos_x = neg_x = pos_y = neg_y = 0;
			for (int i = 0; i + 1 < n; ++i)
			{
				ar[i] = Point(rnd.next(-w * 3 / n - 1, w * 3 / n + 1), rnd.next(-h * 3 / n - 1, h * 3 / n + 1));
				(ar[i].x >= 0 ? pos_x : neg_x) += ar[i].x;
				(ar[i].y >= 0 ? pos_y : neg_y) += ar[i].y;
			}
			if (pos_x - neg_x + abs(pos_x + neg_x) > w * 2 || pos_y - neg_y + abs(pos_y + neg_y) > h * 2)
				continue;
			long long val = abs(pos_x + neg_x) + abs(pos_y + neg_y);
			if (best_val == -1 || val < best_val)
			{
				best_val = val;
				best_ar = ar;
				best_ar.emplace_back(-(pos_x + neg_x), -(pos_y + neg_y));
			}
		}
		sort(best_ar.begin(), best_ar.end());
		pos_x = neg_x = pos_y = neg_y = 0;
		for (int i = 0; i < (int)best_ar.size(); ++i)
		{
			(best_ar[i].x >= 0 ? pos_x : neg_x) += best_ar[i].x;
			(best_ar[i].y >= 0 ? pos_y : neg_y) += best_ar[i].y;
		}
		vector<Point> poly(1);
		poly[0].x = rnd.next(-w - neg_x, w - pos_x);
		poly[0].y = rnd.next(-h - neg_y, h - pos_y);
		for (int i = 0; i < (int)best_ar.size(); ++i)
		{
			if (i && best_ar[i].vct(best_ar[i - 1]) == 0)
				poly.back() = poly.back() + best_ar[i];
			else
				poly.push_back(poly.back() + best_ar[i]);
		}
		poly.pop_back();

		vector<pair<int, int>> res;
		for (auto e : poly)
			res.emplace_back(e.x, e.y);
		return res;
	}
}

namespace gen_polygon
{
	const double PI = acos(-1);

	vector<pair<int, int>> half_main(int argc, char *argv[]) {
		if (argc != 6)
		{
			printf("Usage: gen_polygon [n] [r] [angle] [offset] [seed]\n");
			return {};
		}

		int n = atoi(argv[1]);
		int r = atoi(argv[2]);
		double angle = atoi(argv[3]) * PI / 180;
		int offset = atoi(argv[4]);

		vector<pair<int, int>> res;
		for (int i = 0; i < n; ++i)
			res.emplace_back((int)(r * cos(angle + i * 2 * PI / n) + 0.5) + rnd.next(-offset, offset),
			(int)(r * sin(angle + i * 2 * PI / n) + 0.5) + rnd.next(-offset, offset));
		return res;
	}
}

namespace gen_max
{
	vector<pair<int, int>> ort(vector<pair<int, int>> ar)
	{
		for (int i = 0; i < (int)ar.size(); ++i)
			ar[i] = make_pair(-ar[i].second, ar[i].first);
		return ar;
	}

	vector<pair<int, int>> gen_reverse(int steps)
	{
		vector<pair<int, int>> offset;
		for (int i = 1; i <= steps; ++i)
			offset.emplace_back(1, i);
		offset.emplace_back(0, 2);
		for (int i = steps; i >= 2; --i)
			offset.emplace_back(-1, i);

		auto orted = ort(offset);
		offset.insert(offset.end(), orted.begin(), orted.end());
		orted = ort(orted);
		offset.insert(offset.end(), orted.begin(), orted.end());
		orted = ort(orted);
		offset.insert(offset.end(), orted.begin(), orted.end());
		
		vector<pair<int, int>> res;
		res.reserve(offset.size());
		int x = 0, y = 0;
		for (auto e : offset)
		{
			x += e.first;
			y += e.second;
			res.emplace_back(x, y);
		}
		if (x != 0 || y != 0)
			printf("gen_max::gen_reverse(int steps) did a bad thing");
		return res;
	}

	vector<pair<int, int>> half_main(int argc, char *argv[]) {
		if (argc != 3)
		{
			printf("Usage: gen_max [wh] [seed]\n");
			return {};
		}
		int wh = atoi(argv[1]);

		vector<pair<int, int>> res;
		int l = 0, r = 12501;
		while (l + 1 < r)
		{
			int m = (l + r) / 2;
			res = gen_reverse(m);
			int minx = 0, miny = 0, maxx = 0, maxy = 0;
			for (auto e : res)
			{
				minx = min(minx, e.first);
				miny = min(miny, e.second);
				maxx = max(maxx, e.first);
				maxy = max(maxy, e.second);
			}
			if (maxx - minx > wh * 2 || maxy - miny > wh * 2)
				r = m;
			else
				l = m;
		}
		res = gen_reverse(l);
		int minx = 0, miny = 0;
		for (auto e : res)
		{
			minx = min(minx, e.first);
			miny = min(miny, e.second);
		}
		for (int i = 0; i < (int)res.size(); ++i)
			res[i] = make_pair(res[i].first - minx - wh, res[i].second - miny - wh);
		return res;
	}
}

const int GENERATOR_COUNT = 5;
const string generators[GENERATOR_COUNT] = { "gen_ellipse", "gen_few", "gen_minkowski", "gen_polygon", "gen_max" };
const int arg_count[GENERATOR_COUNT] = { 5, 5, 5, 6, 3 };

vector<pair<int, int>> half_main(int argc, char *argv[])
{
	if (generators[0] == argv[0])
		return gen_ellipse::half_main(argc, argv);
	if (generators[1] == argv[0])
		return gen_few::half_main(argc, argv);
	if (generators[2] == argv[0])
		return gen_minkowski::half_main(argc, argv);
	if (generators[3] == argv[0])
		return gen_polygon::half_main(argc, argv);
	if (generators[4] == argv[0])
		return gen_max::half_main(argc, argv);
	printf("Unknown graph generator %s", argv[0]);
	return {};
}
